//  $Id: prometheus.hh,v 1.41 2004/12/14 18:19:12 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prometheus.hh
//
#ifndef __PROMETHEUS_H__
#define __PROMETHEUS_H__

#include "prom_grid.hh"

//#define PROM_MEASURE_MPATH

/*****************************************************************************
 *                     Prometheus
 ****************************************************************************/
class PromSolver;
class PromMG;
//! class Prometheus: main solver object
/*! 
 * class Prometheus: main solver object
 */
class Prometheus : public PromContext
{  
public:
  Prometheus( const PromComm &comm, PromOptions &opts, 
	      PromPerfMonitor &perf );
  virtual ~Prometheus(); 
  int      Init( bool init1 );
  virtual int Archive( FILE *file = stderr, 
		       const PromArchiveType type = PROM_PRINT);
  //! new level methods
  int        NewLevel( int );
private:
  bool       allDone( PromGrid *lastGrid, const int topMaxEq ) const ;
  int        MakeNewGrid(bool islast, const int topMaxEq, 
			 PromGrid *const lastGrid, PromGrid *& newGrid ) const;
public: 
  int        FinalizeSymbolic();
  int        UpdateGeometry( double * const crds, double * const deltas );
  int        SetMatrixWithNewData( PromMatrix *AA );
  int        CreateFineGrid(const int npLoc, const int lid_ndof[], const int ndf,
			    const PromTable<int> &gid_ghostLid,
			    const PromList<int*> *elemList);
  int        getNumLevels() const;
  int        nBlocksWithGlobalDOF( const int gl ) const;
  int        WritePartitioned();
  int        MakeSplitCoarseGrid(PromGrid *lastGrid, const int nloc, 
				 const int nnodes, PromGrid **newGrid,
				 const int factor ) const;
  int        ConvertToSymm(PromMatrix *const A, const PromOptions &opts)const;
  PromMGType getMGType() const { return mgType_; }

  // Data
  bool          finalized_;
  PromMGType    mgType_;
  PromGrid     *top_grid_;
  PromGrid     *grid0_;
  PromVector   *right_;
  PromVector   *left_;
  PromSolver   *solver_;
  //PromVector   *xVec_; // used in non-fei only!
  //PromVector   *rVec_; // used in non-fei only!
  //PromVector   *bVec_; // used in non-fei only!
  //
  const PromComm &Comm_; 
  PromOptions &options_;
  PromPerfMonitor &perf_mon_;
  PromParentBank parBank_;
}; // Prometheus

/*****************************************************************************
 *                     PromProjKKTSolver
 ****************************************************************************/
//! class PromProjKKTSolver
/*!
 * Krylov methods
 * owns PC object
 */
class PromProjKKTSolver 
{
public:
  PromProjKKTSolver() : projC_(NULL), ksp_mult_work_(NULL), CCt_inv_(NULL), 
			ipiv_(NULL), projLEq_geq_(NULL), workCR_(NULL){}
  ~PromProjKKTSolver(){
    if( projC_ != NULL ) ClearProjC();
    else assert(workCR_ == NULL);
  }
  int ClearProjC(){ 
    assert(projC_ != NULL);
    delete projC_; delete workCR_; delete ksp_mult_work_;
    PetscFree( ipiv_ ); PetscFree( CCt_inv_ ); 
    if( projLEq_geq_ != NULL ) PetscFree( projLEq_geq_ );
    // clear other proj data
    projC_ = NULL; 
    projLEq_geq_ = ipiv_ = NULL; 
    CCt_inv_ = NULL;
    ksp_mult_work_ = workCR_ = NULL; 
    return 0;
  }
  int SetProjC( const PromMatrix * const C,
		PromTable<int> *projCREqs );
  //! projection methods
  int PreSolve( const PromMatrix * const A, const PromVector* const, 
		PromVector * const bb, PromVector * const x_0 ) const;
  int PostSolve( const PromVector * const x_0, PromVector* const xx ) const;
  int ComputeMultiplier( const PromMatrix * const A, 
			 const PromMatrix * const otherC,
			 const PromVector* const xx, PromVector* const pp, 
			 const PromVector * const bb ) const;
  int ApplyProj( const PromVector * const XX, PromVector * const YY ) const;
  int Apply_CCInv_C(const PromVector * const XX, PromVector * const YY) const;
  int Apply_Ct_CCInv(const PromVector * const XX, PromVector*const YY) const;
  int Apply_Ct_add( const PromVector * const GG, PromVector * const XX, 
		    PromVector * const YY ) const;
  int ApplyCCInv( PromVector * const XX ) const;
private:
  int MapProjCRsToAllCRs(const PromVector*const XX,PromVector*const YY )const;
  int MapAllCRsToProjCRs(const PromVector*const GG,PromVector * const YY)const;
public:
  bool isConforming( const PromVector * const XX, double &beta ) const;
  static int SparseDot(const int nz1, const int ip1[], const double vv1[],
		       const int nz2, const int ip2[], const double vv2[],
		       double &dot );
  // data
  const PromMatrix         *projC_;    // C
  mutable PromVector       *ksp_mult_work_;
private:
  double                   *CCt_inv_;  // ( C C' ) ^-1proj
  int                      *ipiv_;     
  int                      *projLEq_geq_;
  mutable PromVector       *workCR_;  
};

/*****************************************************************************
 *                     PromSolver
 ****************************************************************************/
//! class PromSolver - base solver object 
/*!
 * Krylov methods
 * owns PC object
 */
class PromSolver
{
  friend class PromPCMG;
  friend class PromUzawaSolver;
  friend class PromMatrix_base;
  friend class Prometheus_LinSysCore;
public:
  //! normal constructor
  PromSolver( const PromOptions &opts, PromPerfMonitor &perf,
	      const PromMap &map, const PromMap * const pmap = NULL ) : 
    pc_(NULL), A_(NULL), work_(NULL), map_(map), p_map_(pmap), 
    projector_(NULL), alias_(NULL), options_(opts), perf_mon_(perf), 
    rtol_(1.e-5), dtol_(1.e5), atol_(1.e-29),  alloced_(TRUE),
    verbose_(opts.verbose_), nwork_(0), cg_restart_(-300),
    ksp_type_(PROMKSPGMRES), nEig_(0), maxits_(1){} 
  //! shallow copy constructor constructor -- for 'post' smoother
  PromSolver( PromSolver * const solver ) :
    pc_(NULL), A_(NULL), work_(NULL), map_(solver->map_), p_map_(solver->p_map_), 
    projector_(NULL), alias_(solver), options_(solver->options_), 
    perf_mon_(solver->perf_mon_), rtol_(solver->rtol_), dtol_(solver->dtol_),
    atol_(solver->atol_), alloced_(FALSE), verbose_(solver->options_.verbose_),
    nwork_(0), cg_restart_(-300), ksp_type_(PROMNUMKSP), 
    nEig_(0), maxits_(solver->maxits_)
  { 
    assert(solver != NULL && solver->alias_ == NULL); solver->alias_ = this;
  } 
  //! destructor
  ~PromSolver();
  // methods
  int SetPC( PromPC_base *const pc, const PromMatrix_base *AA ){ 
    A_ = AA;
    assert(pc_==NULL); pc_ = pc; // hack for rho solver 
    return 0; 
  }
  int PrintInfo( const int nactivep, const int lev = -1 ) const;
  int SetUp();
  // 
  //! Mult: v2 = (I - P) A (I - P) * v1
  int KSPMult( const PromMatrix_base * const A,
	       const PromVector_base * const XX, 
	       PromVector_base * const YY ) const;
  int SetTolerances(const int maxits, double rtol = -1.,
		    double atol = -1., double dtol = -1. );
  int SetGMRESRestart( int kmax ){ maxits_ = kmax; return 0; }
  // KSP
  int SetNewPVec( const PromMap &map );
  int SetKSPType( PromKSPType type );
  PromKSPType getKSPType() const {
    assert( this != NULL ); 
    if( !alloced_ ) return alias_->getKSPType();
    return ksp_type_;
  }
  int SetKSP( const int maxit, const char* const prestring = NULL,
	      const char* const ksp_string = NULL );
  int SetRTol( double tol ) {
    assert(this!=NULL); // !alloc
    rtol_ = tol;
    return 0;
  } 
  double getRTol() const {
    assert(this!=NULL); // !alloc
    return rtol_;
  } 
  double getATol() const {
    assert(this!=NULL); // !alloc
    return atol_;
  }
  int nEig()const{ return nEig_;}
#undef __FUNCT__
#define __FUNCT__ "PromSolver::SetNEig"
  int SetNEig( int nn ) { 
    if(this==NULL || nn<=0 )SETERRQ(-1,"this==NULL || nn<=0");
    if( origin_.e != NULL ) {
#if defined(PROM_HAVE_COMPLEX)
      PromComplex *pt = dynamic_cast<PromComplex*>(origin_.c);
      if( pt != NULL ) {
	assert(dd_.c[nEig_+1] == -13.0);
	delete [] pt;
      }
      else {
	assert(dd_.e[nEig_+1] == -13.0);
	int ierr = PetscFree( origin_.e );  CHKERRQ(ierr); 
      }
#else
      assert(dd_.e[nEig_+1] == -13.0);
      int ierr = PetscFree( origin_.e );  CHKERRQ(ierr); 
#endif
      origin_.e = d_.e = ee_.e = dd_.e = NULL;
    }
    nEig_ = nn;
    return 0;
  }
  //  
  bool computeSVD() const {
    return ( (nEig() > 0 && ksp_type_ == PROMKSPCG) || 
	     ksp_type_ == PROMKSPGMRES || ksp_type_ == PROMKSPFGMRES );
  }
  int SetComputeSVD();
  int ComputeExtremeSVD( int nits, double &emax, double &emin );
private:
  int ComputeExtremeSVD_cg( int nn, double &emax, double &emin );
  int ComputeExtremeSVD_gmres( int n, double &emax, double &emin );
public:
  // KSP methods
  int SolveGMRES(const PromVector_base * const, 
		 PromVector_base* const,int *its,
		 const bool zeroguess = TRUE,
		 const bool avoidnorms = FALSE, 
		 const PromPCSide = PROM_RIGHT );
private:
  int ResidualGMRES( const PromPCSide pc_side, const PromVector_base *const xx,
		     PromVector_base* const vv_0, 
		     const PromVector_base* const binvf,
		     PromVector_base* const temp );  
  int CycleGMRES( const PromPCSide pc_side, const PromVector_base * const bb, 
		  PromVector_base * const xx, int *reason,
		  int *itcount, double *rnorm0, 
		  const bool avoidnorms );
  int UpdateGMRESHessenberg( const int it, bool hapend, double *res );
  int BuildGMRESSoln( const PromPCSide pc_side, double* nrs, 
		      PromVector_base *vdest,
		      const int it, const PromVector_base *const* vv_0,
		      PromVector_base * const temp);
  int ModifiedGramSchmidtOrthogonalization(const int it);
  int UnmodifiedGramSchmidtOrthogonalization(const int it);
  int TwoUnmodifiedGramSchmidtOrthogonalization(const int it);
#if defined(PROM_HAVE_COMPLEX)
  int UpdateGMRESHessenberg_complex( const int it, bool hapend, double *res );
  int BuildGMRESSoln_complex( const PromPCSide pc_side, PromComplex* nrs, 
			      PromVector_base *vdest,
			      const int it, const PromVector_base *const* vv_0,
			      PromVector_base * const temp);
  int ModifiedGramSchmidtOrthogonalization_complex(const int it);
  int UnmodifiedGramSchmidtOrthogonalization_complex(const int it);
  int TwoUnmodifiedGramSchmidtOrthogonalization_complex(const int it);
#endif
  int PromResidual( const PromPCSide pc_side, const bool zerox,
		    PromVector_base *vsoln, PromVector_base *vt1,
		    PromVector_base *vt2, PromVector_base *vres,
		    PromVector_base *vbinvf, 
		    const PromVector_base * const vb );
  int SandwPre( const double omV[], const int deg, PromPC_base * const pc,
		const PromMatrix_base * const Amat, PromVector_base *const xx,
		PromVector_base *const yy, PromVector_base *const work) const;
  int SandwPost( const double omV[], const int deg, PromPC_base * const pc,
		 const PromMatrix_base * const Amat, PromVector_base *const xx,
		 PromVector_base *const yy, PromVector_base *const work)const;
  int BrezParameters( const PromMatrix_base * const AA,
		      double &mlsOver, double &mlsCf,
		      double &mlsOm, double &mlsOm2 ) ;
public:
  int SolveRichardson(const PromVector_base* const,
		      PromVector_base* const, int *its,
		      const bool zeroguess = TRUE,
		      const bool avoidnorms = FALSE, 
		      const PromPCSide = PROM_RIGHT );
  int SolveBCGS(const PromVector_base* const, 
		PromVector_base* const,int *its,
		const bool zeroguess = TRUE,
		const bool avoidnorms = FALSE, 
		const PromPCSide = PROM_RIGHT ); 
  int cSolveBCGS(const PromVector_base* const, 
		 PromVector_base* const,int *its,
		 const bool zeroguess = TRUE,
		 const bool avoidnorms = FALSE, 
		 const PromPCSide = PROM_RIGHT ); 
  int SolveCR(const PromVector_base*const, PromVector_base* const,int *its,
	      const bool zeroguess = TRUE, const bool avoidnorms = FALSE, 
	      const PromPCSide = PROM_RIGHT );
  int cSolveCR(const PromVector_base*const, PromVector_base* const,int *its,
	       const bool zeroguess = TRUE, const bool avoidnorms = FALSE, 
	       const PromPCSide = PROM_RIGHT );
  int SolveBrez(const PromVector_base* const,
		PromVector_base* const, int *its,
		const bool zeroguess = TRUE,
		const bool avoidnorms = FALSE, 
		const PromPCSide = PROM_RIGHT );
  int SolveCheb(const PromVector_base*const, 
		PromVector_base* const, int *its,
		const bool zeroguess = TRUE,
		const bool avoidnorms = FALSE, 
		const PromPCSide = PROM_RIGHT );
  int cSolveCheb(const PromVector_base*const, 
		 PromVector_base* const, int *its,
		 const bool zeroguess = TRUE,
		 const bool avoidnorms = FALSE, 
		 const PromPCSide = PROM_RIGHT );
  int SolveCG(const PromVector_base* const, 
	      PromVector_base* const, int *its,
	      const bool zeroguess = TRUE,
	      const bool avoidnorms = FALSE, 
	      const PromPCSide = PROM_RIGHT );
  int cSolveCG(const PromVector_base* const, 
	       PromVector_base* const, int *its,
	       const bool zeroguess = TRUE,
	       const bool avoidnorms = FALSE, 
	       const PromPCSide = PROM_RIGHT );
  int SolveTFQMR(const PromVector_base* const, 
		 PromVector_base* const, int *its,
		 const bool zeroguess = TRUE,
		 const bool avoidnorms = FALSE, 
		 const PromPCSide = PROM_LEFT );
  int SolveSYMMLQ(const PromVector_base* const, 
		  PromVector_base* const, int *its,
		  const bool zeroguess = TRUE,
		  const bool avoidnorms = FALSE, 
		  const PromPCSide = PROM_LEFT );
  int SolveMINRES(const PromVector_base* const, 
		  PromVector_base* const, int *its,
		  const bool zeroguess = TRUE,
		  const bool avoidnorms = FALSE, 
		  const PromPCSide = PROM_LEFT );
  int cSolveMINRES( const PromVector_base* const, 
		    PromVector_base* const, int *its,
		    const bool zeroguess = TRUE,
		    const bool avoidnorms = FALSE, 
		    const PromPCSide = PROM_LEFT );
  // PC
  int CreatePC( PromPCType type, const PromVector_base * const v );
  int MakePC( PromGrid * const grid, 
	      int &nblocks, PromIS *&blockISs, bool &kspset,
	      const PromVector *const xx, 
	      const char* const prestring = NULL,
	      const char* const pcname = NULL );
  int SetOperator( const PromMatrix_base * const A );
  int SetNewOperator( const PromMatrix_base * const A );
  int SetOperator_private( const PromMatrix_base * const A );
  //! 
  int PromPCApplyBAorAB( const PromMatrix_base * const Amat,
			 PromPC_base * const pc, 
			 const PromPCSide pc_side,
			 const PromVector_base * const xx,
			 PromVector_base * const dest,
			 PromVector_base * const zz )const;
  //!
  int Solve(const PromVector_base* const bb, 
	    PromVector_base* const xx,
	    int *its, const bool zeroguess = TRUE, 
	    const bool avoidnorms =FALSE);
protected:
  int TestSymmMA( const PromMatrix_base * const Amat,
		  PromPC_base * const pc, PromVector_base *const Z, 
		  PromVector_base *const P, PromVector_base *const R )const;
  int Monitor( const int it, const double dp, const bool forc = FALSE ){
    if(verbose_>1 || (verbose_==1 && it%5==0) || (verbose_>0 && forc)) {
      return PetscPrintf(A_->MPIComm(),"%d (p)KSP Residual norm %12.7e\n",it,dp);
    }
    return 0;
  }
  int converged( double r0, double r ){ 
    assert(this!=NULL); // !alloc
    if( r<r0*rtol_ || r<atol_ ) return 1;
    else if ( r > r0*dtol_) return -1;
    return 0;
  }
public:
  // data 
  static const char * const PromKSPStrings[PROMNUMKSP];
  
  PromPC_base            *pc_;
  const PromMatrix_base  *A_;
  PromVector_base       **work_;
  const PromMap           &map_;
  const PromMap     *const p_map_;
  PromProjKKTSolver       *projector_;
protected:
  PromSolver              *alias_; // !NULL if a smoother 
  const PromOptions       &options_;
  PromPerfMonitor         &perf_mon_;
  float                    rtol_;  // used by aliased 
  float                    dtol_;  // used by aliased 
  float                    atol_;  // used by aliased 
  const bool               alloced_; // is 'alias_' a dependent or am I?
public:
  mutable short int        verbose_;
protected:
  short int                nwork_;
  short int                cg_restart_;
  struct prom_double_complex{
    prom_double_complex(): e(NULL)
#if defined(PROM_HAVE_COMPLEX)  
			 , c(NULL) 
#endif
    {};
    double *e;
#if defined(PROM_HAVE_COMPLEX)  
    PromComplex *c;
#endif
  }origin_, d_, hes_, dd_, ee_;
  
  PromKSPType               ksp_type_;
  short int                 nEig_;       // CG, == nEig_ for GMRES
  short int                 maxits_;     // used by aliased 
public:
  struct prom_history{
    prom_history():helmholtz(0),iters(0),change(-511),level(0),istop(0),isbase(0){}
    unsigned int helmholtz : 1;
    unsigned int iters     : 10;
    int change             : 10;
    unsigned int level     : 4;
    unsigned int istop     : 1;
    unsigned int isbase    : 1;
    unsigned int aux       : 5;
  }history_;
}; // PromSolver

/*****************************************************************************
 *      G-S               PromGSCont
 ****************************************************************************/
typedef enum BlktypeTAG{PROM_INF=0,PROM_INTER,PROM_INF_SUP,PROM_SUP}PromBlockType;
typedef enum BlkListtypeTAG{PROM_INFL=0,PROM_INTER1,PROM_INTER1B,PROM_INF_SUP_WO_HIGHERIS,PROM_INF_SUPL,PROM_INF_SUP_WO_LOWERIS,PROM_INTER2,PROM_INTER2B,PROM_SUPL,PROM_NBLKLISTTYPES}PromBlockListType;

struct PromGSCont{
  PromGSCont():arr_size_(0),list_(NULL){}
  int arr_size_;
  union{
    PromList<int> *list_;
    int *array_;
  };
};

/*****************************************************************************
 *                     PromProcXComm
 ****************************************************************************/
class PromProcXComm 
{ 
public: 
  PromProcXComm( int nn, int neq, int ppart, MPI_Comm cm, const int tg ) :
    part_(ppart),  nNds_(nn),
    nEqs_(neq),    tag_(tg),
    lids_(NULL),   xarr_(NULL),
    req_(MPI_REQUEST_NULL), comm_(cm)
  {
    assert(nNds_>0); 
    int isz=sizeof(int), dsz=sizeof(double), xx=nEqs_*(dsz/isz)+nNds_;
    PetscMalloc((xx+1)*isz,&lids_); assert(lids_ != NULL); 
    lids_[xx] = lids_[xx-1] = -7; 
    xarr_ = (double*)(lids_); 
    xx = nEqs_; while(xx--) xarr_[xx] = 1.e30;
    lids_ += nEqs_ * (dsz/isz); 
    InfBlocks_.arr_size_ = 0;
    InfBlocks_.list_ = new PromList<int>(10);
  }
  ~PromProcXComm() {
    assert(lids_[nNds_-1] != -7); assert(lids_[nNds_] == -7);
    PetscFree( xarr_ );  
    if(InfBlocks_.arr_size_ != 0) PetscFree(InfBlocks_.array_);
  }
  // data 
  const int    part_;
  const int    nNds_;
  const int    nEqs_;
  const int    tag_;
  PromGSCont   InfBlocks_;
  int         *lids_;
  double      *xarr_;
  MPI_Request  req_;
  MPI_Comm     comm_;
};

/*****************************************************************************
 *                     PromBlockXComm
 *****************************************************************************/
class PromBlockXComm
{
public:  
  PromBlockXComm( int nn, int neq, int ppart, const int ggbid ) :
    nEqs_(neq), nNds_(nn),
    part_(ppart),
#if defined(PROM_MEASURE_MPATH)
    path_(-9999), // debug
#endif
    gbid_(ggbid)
  {
    assert(nNds_>0); int isz = sizeof(int);
    PetscMalloc((nNds_+1)*isz,&locBoundIDs_);   assert(locBoundIDs_!=NULL); 
    locBoundIDs_[nNds_] = locBoundIDs_[nNds_-1] = -7;
  }
  ~PromBlockXComm() {
    assert(locBoundIDs_[nNds_-1] != -7 && locBoundIDs_[nNds_] == -7);
    PetscFree(locBoundIDs_);
  }
  // data
  int       *locBoundIDs_;
  const int  nEqs_;
  const int  nNds_; // keep all nodes in agg the same! 
  const int  part_; 
#if defined(PROM_MEASURE_MPATH)
  int path_; // debug
#endif
  const int  gbid_;
};

/*************************************************************
 *                  PromGSBlock
 **************************************************************/
class PromGSBlock 
{
public:
  PromGSBlock( int iidd, int pproc ) : 
#if defined(PROM_MEASURE_MPATH)
    path_(-8888),
#endif
    proc(pproc)
    { 
      bits.done = FALSE;  bits.gid = iidd;
    }
  virtual ~PromGSBlock(){}
  int getProc() const { assert(this!=NULL); return proc; }
  int isMarked() const { assert(this!=NULL); return bits.done; }
  void mark( int b = 1 ){ assert(this!=NULL); bits.done = (b==0) ? 0 : 1; }
  int getGBID() const { assert(this!=NULL); return bits.gid; }
  //void setGBID( int b ){ assert(this!=NULL && b>=0); bits.gid = b; }
#if defined(PROM_MEASURE_MPATH)
  int path_; // debug
#endif
private:
  // data
  struct {
    unsigned int done : 1;
    unsigned int gid : 31;
  }bits;
  const int proc;
};

/*************************************************************
 *                  PromInfSupBlock
 **************************************************************/
class PromInfSupBlock : public PromGSBlock 
{
public:
  PromInfSupBlock( int iidd, int pproc ) : 
    PromGSBlock(iidd,pproc),
    ISBlockDeps_(NULL),    sendXComm_(NULL),
    high_ISBlocks_(10),    low_ISBlocks_(10),
    lowProc_sendXComm_(10),highProc_sendXComm_(10)
    { }
  ~PromInfSupBlock(){
    PromBlockXComm *commx;
    while ( !highProc_sendXComm_.isEmpty() ) {
      commx = highProc_sendXComm_.removeHead();
      delete commx; 
    } 
    while ( !lowProc_sendXComm_.isEmpty() ) {
      commx = lowProc_sendXComm_.removeHead();
      delete commx; 
    }
  } 
  int collectSends( PromTable<PromList<PromBlockXComm*>*> &proc_list );
  //
  PromList<PromGSBlock*> *ISBlockDeps_; // pointer to low(high)Proc_ISBlocks
  PromList<PromBlockXComm*> *sendXComm_; // pointer to low(high)Proc_sendXComm
  // lists of dependents, etc.
  PromList<PromGSBlock*> high_ISBlocks_; // list of PromGSBlock pointers
  PromList<PromGSBlock*> low_ISBlocks_;  // list of PromGSBlock pointers
  // send communicator if proc == myproc, else recv communicator
  PromList<PromBlockXComm*> lowProc_sendXComm_;  // list of PromProcXComm;
  PromList<PromBlockXComm*> highProc_sendXComm_; // list of PromProcXComm;
};

template <class T> class PromAList : public PromList<T>
{
public:
  PromAList(int sz = 5) : PromList<T>(sz), count_(0) {}
  int count_;
};

/*****************************************************************************
 *                     PromGSNode
 ****************************************************************************/
class PromGSNode
{ 
public:
  PromGSNode(int lid, int gid, unsigned int n=0) : b(lid,n), gid_(gid) {}
  int global() const { return gid_; }
  int nb()const{return b.multi_;} 
  // data
  struct T{
    T(int l, unsigned int n):multi_(n),lid_(l){}
    const unsigned int multi_ : 4;
    const int lid_ : 28;
  }b;
private:
  const int gid_;
};

/*****************************************************************************
 *                     PromPCGaussSeidel
 ****************************************************************************/
//! class PromPCGaussSeidel
/*!
  - Parallel Gauss-Seidel Preconditioner
*/
class PromPCGaussSeidel : public PromPC_base
{
  friend class PromSolver;
public: 
  PromPCGaussSeidel( const PromOptions &opt, 
		     PromPerfMonitor &perf );
  int Create( const PromMatrix * const, const PromIS * const domainISs,
	      const int nblocks );
  ~PromPCGaussSeidel();
  // methods
private: 
  int Create_private( const PromIS * const domainISs ); 
public: 
  virtual int Apply_private(const PromVector_base* const bb, 
			    PromVector_base* const xx, const bool zerox=TRUE);
  virtual int SetUp_private();
  virtual const char * getPCString() const {
    return PromPCStrings[PROMPCGAUSSSEIDEL + (symmetric_ ? 1 : 0) ]; 
  }
  virtual const char * getSubPCString() const {
    if( kkt_ != NULL ) return kkt_->getPCString();
    else if( pc_ != NULL ) return pc_->getPCString();
    else return PromPCStrings[PROMPCNONE]; 
  }
  virtual int getNumBlocks() const {
    if( pc_ != NULL ) return pc_->getNumBlocks();
    else if( kkt_ != NULL ) return  kkt_->getNumBlocks();
    else return 0; 
  }
  virtual PromPCType getType() const {
    assert( this != NULL );
    return PROMPCGAUSSSEIDEL;
  }
  virtual int takeNonZeroGuess()const{ return pre_smooth_ ? 1 : -1; }
  virtual int SetNewOperator_private( const PromMatrix_base * const Ab ) {
    assert(this!=NULL); int ierr; assert(Ab!=NULL);
    const PromMatrix *AA = dynamic_cast<const PromMatrix*>(Ab); 
    assert(AA!=NULL);
    if( pc_ != NULL ) {
      ierr = pc_->SetNewOperator( AA );  CHKERRQ(ierr);
    }
    if( A_ != NULL && AA->map_ != A_->map_ )SETERRQ(1,"A->map_ != map_");
    A_ = AA;
    return 0;
  }
  virtual int SetOperator_private( const PromMatrix_base * const Ab ) {
    assert(this!=NULL); int ierr; assert(Ab!=NULL);
    const PromMatrix *AA = dynamic_cast<const PromMatrix*>(Ab); 
    assert(AA!=NULL);
    if( pc_ != NULL ) {
      ierr = pc_->SetOperator( AA );  CHKERRQ(ierr);
    }
    if( A_ != NULL && AA->map_ != A_->map_ )SETERRQ(1,"A->map_ != map_");
    A_ = AA;
    return 0;
  }
private: 
  int Apply_one( const PromVector* const bb, PromVector* const xx,
		 bool zerog = FALSE, bool send_post = TRUE );
  // setup methods
  virtual int Setup_private( PromList<int> ** const block_procs,
			     PromTable<int> **const block_gid_ghostLid,
			     PromList<int> **block_nodeLids,
			     PromList<int> **block_boundLids,
			     PromList<int> **block_adjacBLids,
			     PromTable<PromTable<int>*>*const proc_boundNodes,
			     const PromIS * const domainISs,
			     PromList<int>  *highlist, 
			     PromList<int> *lowlist );
  int ColorProcs(PromList<int> *const highlist, PromList<int> *const lowlist);
  int SetBlockTypes( const PromList<int> *const * const block_procs, 
		     const PromTable<PromTable<int>*>*const proc_bNodes,
		     PromBlockType block_type[], int lid_blockGID[],
		     const PromIS * const domainISs );
  int MakeLists( PromList<int> * const * const block_procs,
		 const PromTable<int> * const * const block_ghosts,
		 const PromBlockType block_type[], const int lid_blockGID[],
		 const PromList<int> * const * const block_nodeLids,
		 const PromList<int> * const * const block_boundNodes,
		 const PromList<int> * const * const block_adajacBlids,
		 const PromIS*const domainISs);
  int SetupNodalGS();
  // solver
  int DoBlocks( const int nb, const int blocks[],
		const PromVector* const bb, 
		PromVector *const xx ) const;
#if defined(PROM_USE_PETSC)
  int DoBlocks_nodal_1(const int nb, const int block[],
		       const PromVector* const b, 
		       PromVector* const x) const; 
  int DoBlocks_nodal_2(const int nb, const int block[],
		       const PromVector* const b, 
		       PromVector* const x) const; 
  int DoBlocks_nodal_3(const int nb, const int block[],
		       const PromVector* const bb,
		       PromVector*const xx) const; 
  int DoBlocks_nodal_6(const int nb, const int block[],
		       const PromVector* const bb, 
		       PromVector* const xx)const; 
  int DoBlocks_nodal(const int nb, const int block[],
		     const PromVector* const bb,PromVector*const xx,
		     const int ndf ) const; 
  int DoBlocks_schwarz_1(const int nb, const int block[],
			 const PromVector* const bb,
			 PromVector* const xx,
			 const PromPC * const pc )const; 
  int DoBlocks_schwarz_2(const int nb, const int block[],
			 const PromVector* const bb,
			 PromVector* const xx,
			 const PromPC * const pc )const;
  int DoBlocks_schwarz_3(const int nb, const int block[],
			 const PromVector* const bb, 
			 PromVector* const xx,
			 const PromPC * const pc )const;
  int DoBlocks_schwarz_6(const int nb, const int block[],
			 const PromVector* const bb, 
			 PromVector* const xx,
			 const PromPC * const pc )const;
  int DoBlocks_schwarz(const int nb, const int block[],
		       const PromVector*const bb,PromVector*const xx,
		       const PromPC * const pc, 
		       const int ndf ) const;
#else
  // petra specific G-S kernals
#endif
  
  // utils
  int GetProc( const int gid, int &proc ) const;
  int PostRecvs( const PromAList<PromProcXComm*> *commxList ) const;
  int SendLocalx(const PromList<PromProcXComm*> *commxList, 
		 const PromVector * const xx)const;
  int RecvDoGhostx( PromAList<PromProcXComm*> *const commxList, 
		    const PromVector* const bb = NULL, 
		    PromVector *const xx = NULL ) const;
  int WaitForList( const PromList<PromProcXComm*> *commxList ) const;
  int SendISblocks( PromTable<PromList<PromBlockXComm*>*> *proc_list, 
		    PromList<MPI_Request*> *smessages,
		    const PromVector *const xx, const int tag ) const;
  int RecvBlockGhostx( double ghostx[], const int nghostEqs,
		       const int tag, const bool iprobe,
		       int &ndone ) const;
public:
  int ReorderNodes( double wieghts[], int len );
private: 
  static int BfsReorder( PromList<int> *blidList, 
			 const PromList<int> *const * const block_adjacBLids, 
			 PromList<int> *list, PromList<int> *nextRoots,
			 bool buff1[], bool buff2[], const bool print = FALSE);
  // data
  int         maxnGSBlocks_; // used for "globalizing" block IDs
  // 'DoBlocks(...)' data
public:
  int         nGSBlocks_;
  int         bs_;
  bool        pre_smooth_;          // hack for alternating up/down smoothin
  bool        symmetric_;
  MPI_Comm    comm_;
  //int        *ghostLid_locEQ_;
  //int        *lid_geq_;
  double     *NodalFactors_;
  double     *ghostx_;      
  int        *NodalFactorPointers_;
  PromPC     *pc_;
  const PromMatrix  *A_;
private: 
  PromTable<PromGSNode*> gbid_ghost_; // gid --> local ghost ID (ie, 0:nghost-1)
  PromGSCont blocksLists_[PROM_NBLKLISTTYPES];// local block IDs each type
  PromList<PromProcXComm*>   lowProc_send_;            // list of X_COMM
  PromAList<PromProcXComm*>  lowProc_recv_;            // list of X_COMM
  PromAList<PromProcXComm*>  highProc_recv_;           // list of X_COMM
  PromList<PromProcXComm*>  highProc_send_;           // list of X_COMM
  PromTable<PromGSBlock*> gbid_ISBlock_; // list of IS_BLOCKS
  PromTable<int>     proc_color_;   // proc --> color
}; // PromPCGaussSeidel

/*****************************************************************************
 *                     PromMG
 ****************************************************************************/
//! class PromMG 
/*!
  - solver stuff for an MG level
*/
class PromMG 
{
  friend class PromPCMG;
friend class Prometheus_LinSysCore;
public:
  PromMG( PromPerfMonitor &perf );
  ~PromMG();
  // methods
  int SetCycles(int nn){ assert(this!=NULL); cycles_ = nn; return 0; }
  int getCycles()const{ assert(this!=NULL); return cycles_; }
  int Create( PromGrid * const grid0, int npre, int npost, 
	      const PromMap *const p_map = NULL,
	      const bool use_iter_on_top = FALSE );
  int SetOperator( const PromMatrix_base * const K ) { 
    A_ = K; int ierr = 0;
    if( presmooth_ != NULL ) ierr = presmooth_->SetOperator( K );
    if( postsmooth_ != NULL ) ierr = postsmooth_->SetOperator( K );
    CHKERRQ(ierr);
    return 0;
  }
  int SetNewOperator( const PromMatrix_base * const K ) { 
    A_ = K; int ierr;
    if( presmooth_ != NULL ) ierr = presmooth_->SetNewOperator(K);
    if( postsmooth_ != NULL ) ierr = postsmooth_->SetNewOperator(K);
    CHKERRQ(ierr);
    return 0;
  }
  int SetNewPVec( const PromMap &map );
  // data
  PromSolver *presmooth_;
  PromSolver *postsmooth_;
  PromVector_base   *b_;
  PromVector_base   *x_;
  PromVector_base   *r_;
  const PromMatrix_base   *interpolate_;  // with next (up) grid
  const PromMatrix_base   *restrict_;
  const PromMatrix_base   *A_;
  PromMG       *prev_;
  PromMG       *next_;
  int           factor_;
protected:
  PromPerfMonitor &perf_mon_;
  int    cycles_;
}; // PromMG

/*****************************************************************************
 *                     PromPCMG
 ****************************************************************************/
//! class PromPCMG
/*!
  - Multigrid Preconditioner
*/
class PromPCMG : public PromPC_base
{
public:
  PromPCMG( const PromOptions &opts, PromPerfMonitor &perf ) :
    PromPC_base( opts, perf ), mgstack_(NULL) {}
  virtual ~PromPCMG();
  // methods
  PromPCType getType() const {
    assert( this != NULL );
    return PROMPCMG;
  }
  // applies PC part in shell PC
  int Apply_private( const PromVector_base* const bb,
		     PromVector_base* const xx ,const bool zerox = TRUE ); 
  virtual int SetUp_private();
  const char * getPCString() const {
    return PromPCStrings[PROMPCMG]; 
  }
  const char * getSubPCString() const {
    if( kkt_ != NULL ) return kkt_->getPCString();
    return PromPCStrings[PROMPCNONE]; 
  }
  virtual int SetOperator_private( const PromMatrix_base * const K ) {
    assert(this!=NULL); int ierr;   
    if( mgstack_ != NULL ) {
      ierr = mgstack_->SetOperator( K );  CHKERRQ(ierr);
    }
    return 0;
  }
  virtual int SetNewOperator_private( const PromMatrix_base * const K ) {
    assert(this!=NULL); int ierr;
    if( mgstack_ != NULL ) {
      ierr = mgstack_->SetNewOperator( K );  CHKERRQ(ierr);
    }
    return 0;
  } 
  int PrintLevelInfo() const;
  int Create( PromGrid * const grid0, const PromMap *p_map = NULL );
private:
  int ACycle( PromMG * const mg, const bool zerox = TRUE );
  int MCycle( PromMG * const mg, int level, const bool zerox = TRUE );
  int FCycle( PromMG * const mg, const bool zerox = TRUE );
public:
  // data
  PromMG* mgstack_;
  PromMGAlgo algo_;
}; // PromPCMG

/*****************************************************************************
 *                     PromPCNodalASM
 ****************************************************************************/
//! class PromPCNodalASM
/*!
  - Nodal block Additive Schwarz preconditioner
*/
class PromPCNodalASM : public PromPC_base
{
public:
  //! constructor
  PromPCNodalASM( const PromOptions &opts, PromPerfMonitor &perf )
    : PromPC_base(opts,perf),invBlocks_(NULL),r_(NULL),work_(NULL),A_(NULL){}
  //! destructor
  ~PromPCNodalASM(){ 
    if( invBlocks_ ) PetscFree( invBlocks_ );
    if( r_ ) delete r_;
    if( work_ ) delete work_; //PromVector::Delete( work_ ); 
  }
  //! methods
  virtual PromPCType getType() const {
    assert( this != NULL );
    return PROMPCNODALASM;
  }
  virtual const char * getPCString() const {
    return PromPCStrings[PROMPCNODALASM]; 
  }
  virtual const char * getSubPCString() const {
    if( kkt_ != NULL ) return kkt_->getPCString();
    return PromPCStrings[PROMPCNONE];
  }
  //! SetOperator -- set operator and flag for 'need-factorize'
  virtual int SetOperator_private( const PromMatrix_base * const AA_b ) {
    assert(this!=NULL); 
    if( AA_b == NULL ) SETERRQ(1,"AA_b == NULL");
    const PromMatrix * const AA = dynamic_cast<const PromMatrix*>(AA_b);
    if(AA==NULL)SETERRQ(1,"A not a PromMatrix");
    A_ = AA;
    return 0;
  } 
  //! SetNewOperator -- set operator and flag for 'needs-resetup'
  virtual int SetNewOperator_private( const PromMatrix_base * const AA_b ) {
    assert(this!=NULL); 
    if( AA_b == NULL ) SETERRQ(1,"AA_b == NULL");
    const PromMatrix * const AA = dynamic_cast<const PromMatrix*>(AA_b);
    if(AA==NULL)SETERRQ(1,"A not a PromMatrix");
    A_ = AA;

    return 0;
  }
  //! Construction that needs grid!
  int Create(); 
  int SetUp_private();
  virtual int GetOperator( const PromMatrix **K_out,
			   PromVector **r_out, const PromVector_base*const XX)
    {
      if( work_ == NULL ) {
	work_ = new PromVector( XX ); assert(r_==NULL); 
	r_ = new PromVector( XX );
      } 
      *K_out = A_; *r_out = work_;
      return 0;
    }
  //! applies PC part in shell PC
  virtual int Apply_private(const PromVector_base* const bb,
			    PromVector_base* const xx,const bool zerox = TRUE);
private:
  int Apply_one( const PromVector* const bb, PromVector* const xx );
  // data
  double           *invBlocks_; 
  PromVector       *r_; 
  PromVector       *work_; 
  const PromMatrix *A_; 
}; // PromPCNodalASM

#endif //__PROMETHEUS_H__
